#!/usr/bin/perl -w
#
# use:
# gatttool -b 88:6B:XX:XX:XX:XX --char-write-req -a 0x0009 -n 0300 --listen | ./121gw.pl
#
# to find the MAC of your 121gw enable bluetooth on the 121gw and use:
# hcitool lescan
#
# press crtl-C when finished
#

use strict;
use Time::HiRes qw( gettimeofday );
use utf8;
use feature 'unicode_strings';
use open ':std', ':encoding(UTF-8)';
$|=1;

# our input buffer (array)
my @input = ();

# 121gw mode definitions
my @mode = ();
$mode[  0] = "Low_Z";		$mode[  1] = "DCV";				$mode[  2] = "ACV";			$mode[  3] = "DCmV";
$mode[  4] = "ACmV";		$mode[  5] = "Temp";			$mode[  6] = "Hz";			$mode[  7] = "mS";
$mode[  8] = "Duty";		$mode[  9] = "Resistor";		$mode[ 10] = "Continuity";	$mode[ 11] = "Diode";
$mode[ 12] = "Capacitor";	$mode[ 13] = "ACuVA";			$mode[ 14] = "ACmVA";		$mode[ 15] = "ACVA";
$mode[ 16] = "ACuA";		$mode[ 17] = "DCuA";			$mode[ 18] = "ACmA";		$mode[ 19] = "DCmA";
$mode[ 20] = "ACA";			$mode[ 21] = "DCA";				$mode[ 22] = "DCuVA";		$mode[ 23] = "DCmVA";
$mode[ 24] = "DCVA";		$mode[100] = "_TempC";			$mode[105] = "_TempF";		$mode[110] = "_Battery";
$mode[120] = "_APO_On";		$mode[125] = "_APO_Off";		$mode[130] = "_YEAR";		$mode[135] = "_DATE";
$mode[140] = "_TIME";		$mode[150] = "_BURDEN_VOLTAGE";	$mode[160] = "_LCD";		$mode[180] = "_dBm";
$mode[190] = "_Interval";

# 121gw range definitions
my @range = (
	["V",     "Voltage LowZ (V)",    [4,0,0,0,0,0,0], "       " ],    #0
	["V",     "Voltage DC (V)",      [1,2,3,4,0,0,0], "       " ],    #1
	["V",     "Voltage AC (V)",      [1,2,3,4,0,0,0], "       " ],    #2
	["V",     "Voltage DC (mV)",     [2,3,0,0,0,0,0], "mm     " ],    #3 
	["V",     "Voltage AC (mV)",     [2,3,0,0,0,0,0], "mm     " ],    #4
	["°C",    "Temp (°C)",           [4,0,0,0,0,0,0], "       " ],    #5
	["Hz",    "Frequency (Hz)",      [2,3,1,2,3,0,0], "  kkk  " ],    #6
	["ms",    "Period (s)",          [1,2,3,0,0,0,0], "       " ],    #7
	["%",     "Duty (%)",            [4,0,0,0,0,0,0], "       " ],    #8
	["Ω",     "Resistance (Ω)",      [2,3,1,2,3,1,2], "  kkkMM" ],    #9
	["Ω",     "Continuity (Ω)",      [3,0,0,0,0,0,0], "       " ],    #10
	["V",     "Diode (V)",           [1,2,0,0,0,0,0], "       " ],    #11
	["F",     "Capacitance (F)",     [3,4,2,3,4,5,0], "nnuuuu " ],    #12
	["VA",    "Power AC (uVA)",      [3,4,4,5,0,0,0], "uuuu   " ],    #13
	["VA",    "Power AC (mVA)",      [4,5,2,3,0,0,0], "mm     " ],    #14
	["VA",    "Power AC (VA)",       [4,5,2,3,0,0,0], "mm     " ],    #15
	["A",     "Current AC (uA)",     [2,3,0,0,0,0,0], "uu     " ],    #16
	["A",     "Current DC (uA)",     [2,3,0,0,0,0,0], "uu     " ],    #17
	["A",     "Current AC (mA)",     [3,1,2,0,0,0,0], "mmm    " ],    #18
	["A",     "Current DC (mA)",     [1,2,0,0,0,0,0], "mm     " ],    #19
	["A",     "Current AC (A)",      [3,1,2,0,0,0,0], "m      " ],    #20
	["A",     "Current DC (A)",      [3,1,2,0,0,0,0], "m      " ],    #21
	["VA",    "Power DC (uVA)",      [3,4,4,5,0,0,0], "uuuu   " ],    #22
	["VA",    "Power DC (mVA)",      [2,3,3,4,0,0,0], "mmmm   " ],    #23
	["VA",    "Power DC (VA)",       [4,5,2,3,0,0,0], "mm     " ],    #24
);


sub time_stamp() {
	my ( $s, $us ) = gettimeofday;
	return sprintf "%d.%03d", $s, $us / 1000;
}

sub print_error(@) {
	#return;
	print STDERR time_stamp(), " ", @_;
}

sub print_data(%) {
	my %rec = @_;

	my $value;
	my $format;
	my $str = "";

	# year, month and serial number
	$str .= sprintf "%04d-%02d ", $rec{year} + 2000, $rec{month};
	$str .= sprintf "%05d ", $rec{serial};
	delete($rec{year}); delete($rec{month}); delete($rec{serial});

	$str .= sprintf "%16s ", $range[$rec{mMode}][1];
	# main display
	if ( $rec{mOvl} == 0 ) {
		my $unit = $range[$rec{mMode}][0];
		my $pref = substr($range[$rec{mMode}][3],$rec{mRidx},1);
		my $prec = $range[$rec{mMode}][2][$rec{mRidx}];
		$unit = $pref . $unit;
		$unit =~ s/ //;
		
		$value  = $rec{mValue} / ( 10 ** (5 - $prec) );
		if ( $rec{mSign} > 0 ) {
			$value = 0 - $value;
		}
		$format = sprintf "%%7.%df", ( 5 - $prec );
		$str .= sprintf "$format %-3s", $value, $unit
	} else {
		$str .= sprintf "   -OVL-   ";
	}
	delete($rec{mMode}); delete($rec{mRidx}); delete($rec{mValue}); delete($rec{mSign}); delete($rec{mOvl});

	# secondary display
	if ( $rec{sOvl} == 0 ) {
		$value  = $rec{sValue} / ( 10 ** $rec{sDec});
		if ( $rec{sSign} > 0 ) {
			$value = 0 - $value;
		}
		$format = sprintf "%%7.%df", $rec{sDec};
		if ( length($mode[$rec{sMode}]) ) {
			$str .= sprintf "%10s $format ", $mode[$rec{sMode}], $value;
		}
	} else {
		$str .= sprintf "   -OVL-   ";
	}
	delete($rec{sMode}); delete($rec{sDec}); delete($rec{sValue}); delete($rec{sSign}); delete($rec{sOvl});

	print STDOUT time_stamp , " ", $str, "\n";
}

sub hexdigits_to_value_dec(@) {
	my @bytes = @_;
	my $value = 0;

	foreach my $byte (@bytes) {
		$value *= 10;
		if ( $byte >= 0x30 && $byte <= 0x39 ) {
			$value += $byte - 0x30;
		} else {
			print_error "hexdigits_to_value_dec: Invalid data\n";
			return -1;
		}
	}
	return $value;
}

sub hexdigits_to_value_hex(@) {
	my @bytes = @_;
	my $value = 0;

	foreach my $byte (@bytes) {
		$value <<= 4;
		if ( $byte >= 0x30 && $byte <= 0x39 ) {
			$byte -= 0x30;
		} elsif ( $byte >= 0x41 && $byte <= 0x46 ) {
			$byte -= 0x37;
		} elsif ( $byte >= 0x61 && $byte <= 0x66 ) {
			$byte -= 0x57;
		} else {
			return -1;
		}
		$value += $byte;
	}
	return $value;
}

sub process_record(@) {
	my @values = @_;
	my $value;

	# We received an array of 38 2-character strings
	# e.g.: values[0] = "31", values[1] = "37", values[2] = "30", ...
	# The protocol appears to be quite braindead with beginning codes as 
	# decimal digits and the rest in hexadezimal digits. So we convert our 
	# array of strings representing a byte to an array of half bytes (nibbles)

	# now we use a hash to store parsed values
	my %record = ();

	# byte 0,1: two bytes contain the year as two
	# decimal coded ASCII digits (but it's wrong, it's always 17)
	$value = hexdigits_to_value_dec(splice(@values,0,2));
	return unless ($value >= 0);
	$record{year} = $value;

	# byte 2,3: two bytes contain the month as two
	# decimal coded ASCII digits (but it's wrong, it's always 08)
	$value = hexdigits_to_value_dec(splice(@values,0,2));
	return unless ($value >= 0);
	$record{month} = $value;

	# byte 4,5,6,7,8: Now follow 5 bytes as serial number, they seem always zero 
	$value = hexdigits_to_value_dec(splice(@values,0,5));
	return unless ($value >= 0);
	$record{serial} = $value;


	# The following data to the end of the packet is hexadecimal coded digits
	
	# byte 9,10: two bytes for the mode of the meter
	$value = hexdigits_to_value_hex(splice(@values,0,2));
	$record{mMode}  = $value;

	# byte 11: overload and pos/neg inidication
	$value = hexdigits_to_value_hex(splice(@values,0,1));
	$record{mSign} = 0;
	if ( $value & 0x04 ) {
		$record{mSign} = 1;
	}
	$record{mOvl} = 0;
	if ( $value & 0x08 ) {
		$record{mOvl} = 1;
	}

	# byte 12: Main range index to lookup for SI prefix and precision
	$value = hexdigits_to_value_hex(splice(@values,0,1));
	$record{mRidx} = $value;

	# byte 13,14,15,16 main display value
	$value = hexdigits_to_value_hex(splice(@values,0,4));
	$record{mValue} = $value;

	# byte 17,18: submode 
	$value = hexdigits_to_value_hex(splice(@values,0,2));
	$record{sMode} = $value;

	# byte 19: submode overload, sign, Hz and K indication
	$value = hexdigits_to_value_hex(splice(@values,0,1));

	$record{sHz} = 0;
	if ( $value & 0x01 ) {
		$record{sHz} = 1;
	}

	$record{sK} = 0;
	if ( $value & 0x02 ) {
		$record{sK} = 1;
	}

	$record{sSign} = 0;
	if ( $value & 0x04 ) {
		$record{sSign} = 1;
	}

	$record{sOvl} = 0;
	if ( $value & 0x04 ) {
		$record{sOvl} = 1;
	}

	# byte 20: submode decimal point
	$value = hexdigits_to_value_hex(splice(@values,0,1));
	$record{sDec} = $value;

	# byte 21,22,23,24: value of the small display
	$value = hexdigits_to_value_hex(splice(@values,0,4));
	$record{sValue} = $value;

	# byte 25: bar graph enabled
	$value = hexdigits_to_value_hex(splice(@values,0,1));

	# byte 26: bar range 0 - 150
	$value = hexdigits_to_value_hex(splice(@values,0,1));

	# byte 27,28: bar graph value
	$value = hexdigits_to_value_hex(splice(@values,0,2));

	# byte 29,30: status bits 1
	$value = hexdigits_to_value_hex(splice(@values,0,2));

	# byte 31,32: status bits 2
	$value = hexdigits_to_value_hex(splice(@values,0,2));

	# byte 33,34: status bits 3
	$value = hexdigits_to_value_hex(splice(@values,0,2));

	# byte 35: unknown
	$value = hexdigits_to_value_hex(splice(@values,0,1));

	# byte 36: unknown
	$value = hexdigits_to_value_hex(splice(@values,0,1));

	# byte 37: unknown
	$value = hexdigits_to_value_hex(splice(@values,0,1));

	return(%record);
}


#
# This is the main loop reading lines from gatttool on STDIN
#
# we just filter out the data bytes from the gatttool output
#
while(<>) {
	#
	# This is the reading section
	#
	# We just read from STDIN, grab the characters we want and push them 
	# into an array. A line looks like this:
	# Indication   handle = 0x0008 value: 33 42 46 30 31 30 33 31 42 43 33 30 30 30 31 30 34 34 30 30
	# with "(.*)" we match against the rest of the line, it is placed in variable $1
	if ( /^Indication   handle = 0x0008 value: (.*)/ ) {

		# split the line on every space char into an array of 2 char strings
		my @in = split(' ',$1);

		# push the values interpreted as hex bytes into our input buffer
		foreach my $byte (@in) {
			push(@input, hex($byte));
		}
	}

	#
	# Now we have a look at the data
	#
	# Every record starts with "0xf2" so search for it in our @input array
	# and remove the byte until we find 0xf2 or the array is empty
	while ( $#input > 0 &&  $input[0] != 0xf2 ) {
		# This removes the first element in @input
		shift(@input);
	}

	# if we have 38 or more bytes in our array walk through it
	# and try to find 38 bytes without 0xf2. If we find 0xf2 the 
	# record is incomplete
	if ( $#input >= 38 ) {

		# remove the leading 0xf2
		shift(@input);

		# use an empty array for our record
		my @record = ();

		# remove the next 37 values from bytes and place them into @record
		# and check that there is no other 0xf2 in between as this would
		# indicate a failed transfer
		my $read = 0;
		while ( $#input > 0 && $#record < 37 && $input[0] != 0xf2 ) {
			my $byte = shift(@input);
			$read++;
			if ( $byte >= 0x30 && $byte <= 0x39 ) {
				# check if it's a valid decimal digit
				push(@record,$byte);
			} elsif ( $byte >= 0x41 && $byte <= 0x46 ) {
				# check if it's a char A - F
				push(@record,$byte);
			} else {
				print_error sprintf("invalid byte: 0x%02x on pos %d\n", $byte, $read);
				@record = ();
				while ( $#input > 0 && $input[0] != 0xf2 ) {
					shift(@input);
				}
			}
		}

		# if we found 37 bytes wihtout another 0xf2, then process the record
		if ( $#record == 37 ) {
			my %record = process_record(@record);
			if (%record) {
				print_data(%record);
			}
		}

		# else show that we got an error
		elsif ( $#record > 0 ) {
			print_error "incomplete record\n";
		}
	}
}

# vim: ts=4 ai
